//------------------------------------------------------------------------------
// Digitally controlled 2.1 channel power amplifier system.
// PIC18F452 (IC1) firmware.
// Dilshan R Jayakody <jayakody2000lk@gmail.com>
//
// This Source Code Form is subject to the terms of the 
// Mozilla Public License, v. 2.0. If a copy of the MPL was not distributed 
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
//------------------------------------------------------------------------------

#include "subamp452.h"

unsigned char rotary_value, last_rotary_value, audio_mute = FALSE;
unsigned char vol_lvl = DEFAULT_VOLUME, bass_lvl = DEFAULT_BASS, trib_lvl = DEFAULT_TRIBBLE;
unsigned char loudness = DEFAULT_LOUDNESS, bal_lvl = DEFAULT_BALANCE;
unsigned char bass_add = DEFAULT_BASS_ADD, base_gain = DEFAULT_BASE_GAIN, base_param = 0x00, sys_state = 0x00;
unsigned int timeout_counter = TIMEOUT_LIMIT;

// main entry point.
void main()
{
  init_system();
  rotary_value = 0;
  draw_homescreen();
}

// TDA7315 sound processor control routine.
void update_tone_control(unsigned char tone_command)
{
  I2C1_Start();
  I2C1_Wr(0x80);
  I2C1_Wr(tone_command);
  I2C1_Stop();
  return;
}

// LA2650 bass processor control routine.
void update_bass_processor(unsigned char bass_command)
{
  unsigned char port_b_state, pos = 8;
  while(pos > 0)
  {
    port_b_state = (0x1F & PORTB);
    port_b_state |= (bass_command & (1 << (pos -1))) > 0 ? 0x40 : 0x00;
    PORTB = port_b_state;
    Delay_ms(5);
    
    port_b_state |= 0x20;
    PORTB = port_b_state;
    Delay_ms(5);
    
    pos--;
  }
  
  PORTB = (PORTB & 0x1F) | 0xA0;
  return;
}

// audio mute control routine.
void update_mute_state(unsigned char is_mute)
{
  update_tone_control((is_mute == TRUE) ? 0x9F : 0x80);
  asm nop;
  update_tone_control((is_mute == TRUE) ? 0xBF : 0xA0);
}

void draw_system_volume()
{
  if(audio_mute == TRUE)
    Lcd_Out(1, 1, "   Audio Mute   ");
  else
  {
    Lcd_Out(1, 1, "Volume: ");
    lcd_print_value(1, 9, (MAX_VOLUME_LIMIT - vol_lvl));
  }
}

// print given value on specified row and column.
void lcd_print_value(char row, char col, unsigned char value)
{
  char output_buffer[9], scan_pos = 0;
  ByteToStr(value, output_buffer);
  while(scan_pos < 9)
  {
    // fill remaning buffer space with whitespaces to avoid overlaps.
    if(scan_pos >= 0x03)
      output_buffer[scan_pos] = 0x20;
    scan_pos++;
  }
  
  Lcd_Out(row, col, output_buffer);
  return;
}

void draw_homescreen()
{
  unsigned char portb_state = (PORTB & 0xFC), tmp_mute;

  timeout_counter = TIMEOUT_LIMIT;
  rotary_value = last_rotary_value = vol_lvl;
  Lcd_Cmd(_LCD_CLEAR);

  while(1)
  {
    if(((PORTB & 0x08) == 0x08) && ((portb_state & 0x08) == 0x00))
    {
      // system power ON/OFF control.
      if((sys_state & 0x01) == 0x01)
      {
        // shutdown amplifier system.
        update_mute_state(TRUE);
        Delay_ms(100);
        
        PORTC = 0x00;
        sys_state = 0x00;
        
        Lcd_Cmd(_LCD_CLEAR);
      }
      else
      {
        // powerup amplifier system.
        update_mute_state(TRUE);
        Delay_ms(100);
        
        PORTC = 0x03;
        timeout_counter = TIMEOUT_LIMIT;
        sys_state = 0x01;
        Delay_ms(500);
        
        audio_mute = FALSE;
        update_mute_state(audio_mute);
        draw_system_volume();
      }

      Delay_ms(50);
    }

    if((sys_state & 0x01) ==  0x01)
    {
      if(((PORTB & 0x04) == 0x04) && ((portb_state & 0x04) == 0x00))
      {
        // execute system setup menu.
        timeout_counter = TIMEOUT_LIMIT;
        PORTC |= 0x02;
        sys_state &= 0xFD;
        Delay_ms(10);
        
        draw_setupscreen();
        // exit from system setup menu.
        timeout_counter = TIMEOUT_LIMIT;
        Lcd_Cmd(_LCD_CLEAR);
        draw_system_volume();
      }

      if(((PORTB & 0x10) == 0x10) && ((portb_state & 0x10) == 0x00))
      {
        // audio mute activate / deactivate routine.
        audio_mute = ~audio_mute;
        update_mute_state(audio_mute);
        PORTC |= 0x02;
        timeout_counter = TIMEOUT_LIMIT;
        sys_state &= 0xFD;
        Delay_ms(10);
        
        Lcd_Cmd(_LCD_CLEAR);
        draw_system_volume();
      }

      if(rotary_value != last_rotary_value)
      {
        PORTC |= 0x02;
        sys_state &= 0xFD;
        timeout_counter = TIMEOUT_LIMIT;

        // change master volume level.
        if(rotary_value > MAX_VOLUME_LIMIT)
          rotary_value = MAX_VOLUME_LIMIT;

        update_tone_control(rotary_value & MAX_VOLUME_LIMIT);
        last_rotary_value = rotary_value;
        vol_lvl = rotary_value;
        EEPROM_Write(0x00, vol_lvl);
        draw_system_volume();
      }

    }

    // check for home screen timeout limit to shutdown the display.
    if(timeout_counter > 0)
    {
      if((--timeout_counter) == 0x00)
      {
        Lcd_Cmd(_LCD_CLEAR);
        PORTC &= 0xFD;
        sys_state |= 0x02;
      }
    }

    portb_state = (PORTB & 0xFC);
    Delay_ms(10);
  }

  return;
}

void draw_setupscreen()
{
  unsigned char portb_state = (PORTB & 0xFC);
  unsigned char is_in_menu = TRUE, menu_pos, menu_control = 0x1;

  timeout_counter = TIMEOUT_LIMIT;
  Lcd_Cmd(_LCD_CLEAR);

  // reset menu position to home location.
  menu_pos = rotary_value = 0;
  last_rotary_value = rotary_value + 1;

  Lcd_Out(1, 1, "> System Options");

  while(menu_control)
  {
    // system power OFF control.
    if(((PORTB & 0x08) == 0x08) && ((portb_state & 0x08) == 0x00))
    {
      // shutdown amplifier system.
      update_mute_state(TRUE);
      Delay_ms(100);
      
      PORTC = 0x00;
      sys_state = 0x00;
      Lcd_Cmd(_LCD_CLEAR);
      break;
    }

    // audio mute activate / deactivate routine.
    if(((PORTB & 0x10) == 0x10) && ((portb_state & 0x10) == 0x00))
    {
      timeout_counter = TIMEOUT_LIMIT;
      audio_mute = ~audio_mute;
      update_mute_state(audio_mute);
      Delay_ms(20);
    }

    // disable global interrupts to avoid sync issues.
    INTCON &= 0x7F;
    // check rotary encode has any change.
    if(last_rotary_value != rotary_value)
    {
      timeout_counter = TIMEOUT_LIMIT;

      if(is_in_menu == TRUE)
      {
        if(rotary_value > 8)
          rotary_value = 8;

        // update LCD main menu position based on rotary encoder position.
        switch(rotary_value)
        {
          case 0:
            Lcd_Out(2, 1, "  Bass:    ");
            lcd_print_value(2, 8, bass_lvl);
            Lcd_Out(2, 11, "     ");
            break;
          case 1:
            Lcd_Out(2, 1, "  Tribble: ");
            lcd_print_value(2, 11, trib_lvl);
            break;
          case 2:
            Lcd_Out(2, 1, "  Loudness: ");
            Lcd_Out(2, 13, (loudness == 0x00) ? "On " : "Off ");
            break;
          case 3:
            Lcd_Out(2, 1, "  Balance: ");
            switch(bal_lvl)
            {
              case 0x00:
                Lcd_Out(2, 12, "L      ");
                break;
              case 0x02:
                Lcd_Out(2, 12, "R      ");
                break;
              default:
                Lcd_Out(2, 12, "L+R    ");
            }
            break;
          case 4:
            Lcd_Out(2, 1, "  Subwoofer: ");
            Lcd_Out(2, 14, ((base_param & 0x80) == 0x80) ? "On " : "Off ");
            break;
          case 5:
            Lcd_Out(2, 1, "  Bass Mix: ");
            Lcd_Out(2, 13, ((base_param & 0x20) == 0x20) ? "On " : "Off ");
            break;
          case 6:
            Lcd_Out(2, 1, "  Sub Gain: ");
            lcd_print_value(2, 13, base_gain);
            break;
          case 7:
            Lcd_Out(2, 1, "  Base Add: ");
            lcd_print_value(2, 13, bass_add);
            break;
          case 8:
            Lcd_Out(2, 1, "  Return       ");
            break;
        }

        menu_pos = rotary_value;
      }
      else
      {
        // set new tone/effect value of the selected option.
        switch(menu_pos)
        {
          // set bass level.
          case 0:
            // limit maximum bass value limit.
            if(rotary_value > 0x0F)
              rotary_value = 0x0F;
            update_tone_control(0x60 | ((rotary_value > 7) ? (rotary_value & 0x08) | ((~rotary_value) & 0x07) : rotary_value));
            bass_lvl = rotary_value;
            lcd_print_value(2, 8, bass_lvl);
            EEPROM_Write(0x01, bass_lvl);
            break;
          // set tribble level.
          case 1:
            // limit maximum tribble value limit.
            if(rotary_value > 0x0F)
              rotary_value = 0x0F;
            update_tone_control(0x70 | ((rotary_value > 7) ? (rotary_value & 0x08) | ((~rotary_value) & 0x07) : rotary_value));
            trib_lvl = rotary_value;
            EEPROM_Write(0x02, trib_lvl);
            lcd_print_value(2, 11, trib_lvl);
            break;
          // set loudness effect.
          case 2:
            loudness = (loudness > 0) ? 0x00 : 0x04;
            update_tone_control(0x40 | loudness);
            if(rotary_value == 0xFF)
              rotary_value = 0;
            EEPROM_Write(0x03, loudness);
            Lcd_Out(2, 13, (loudness == 0x00) ? "On " : "Off ");
            break;
          // set speaker balance levels.
          case 3:
            if(rotary_value > 0x03)
              rotary_value = 0x03;
            switch(rotary_value)
            {
              // activate left speaker
              case 0:
                update_tone_control(0x80);
                asm nop;
                update_tone_control(0xBF);
                Lcd_Out(2, 12, "L      ");
                break;
              // activate right speaker.
              case 2:
                update_tone_control(0xA0);
                asm nop;
                update_tone_control(0x9F);
                Lcd_Out(2, 12 , "R     ");
                break;
              // activate both speakers.
              default:
                update_tone_control(0xA0);
                asm nop;
                update_tone_control(0x80);
                Lcd_Out(2, 12, "L+R    ");
            }
            bal_lvl = rotary_value;
            EEPROM_Write(0x04, bal_lvl);
            break;
          // subwoofer base no/off control.
          case 4:
            base_param = ((base_param & 0x80) == 0x80) ? (base_param & 0x7F) : (base_param | 0x80);
            update_bass_processor(base_param | bass_add | (base_gain << 3));
            if(rotary_value == 0xFF)
              rotary_value = 0;
            EEPROM_Write(0x07, base_param);
            Lcd_Out(2, 14, ((base_param & 0x80) == 0x80) ? "On " : "Off ");
            break;
          // on/off bass mix to left and right channels.
          case 5:
            base_param = ((base_param & 0x20) == 0x20) ? (base_param & 0xDF) : (base_param | 0x20);
            update_bass_processor(base_param | bass_add | (base_gain << 3));
            if(rotary_value == 0xFF)
              rotary_value = 0;
            EEPROM_Write(0x07, base_param);
            Lcd_Out(2, 13, ((base_param & 0x20) == 0x20) ? "On " : "Off ");
            break;
          // subwoofer base level control.
          case 6:
            // limit maximum subwoofer bass level.
            if(rotary_value > 0x03)
              rotary_value = 0x03;
            update_bass_processor(base_param | bass_add | (rotary_value << 3));
            lcd_print_value(2, 13, rotary_value);
            base_gain = rotary_value;
            EEPROM_Write(0x06, base_gain);
            break;
          // base level to subwoofer amplifier stage.
          case 7:
            // limit maximum subwoofer bass level limit.
            if(rotary_value > 0x07)
              rotary_value = 0x07;
            update_bass_processor(base_param | rotary_value | (base_gain << 3));
            lcd_print_value(2, 13, rotary_value);
            bass_add = rotary_value;
            EEPROM_Write(0x05, bass_add);
            break;
        }
      }

      last_rotary_value = rotary_value;
    }

    // enter / leave menu option.
    if(((PORTB & 0x04) == 0x04) && ((portb_state & 0x04) == 0x00))
    {
      timeout_counter = TIMEOUT_LIMIT;

      if(is_in_menu == TRUE)
      {
        // main menu item is selected.
        is_in_menu = FALSE;
        switch(menu_pos)
        {
          case 0:
            rotary_value = bass_lvl;
            break;
          case 1:
            rotary_value = trib_lvl;
            break;
          case 4:
          case 5:
          case 2:
            rotary_value = 0x04;
            break;
          case 3:
            rotary_value = bal_lvl;
            break;
          case 6:
            rotary_value = base_gain;
            break;
          case 7:
            rotary_value = bass_add;
            break;
          case 8:
            // exit from setup menu.
            menu_control = 0;
            break;
        }

        last_rotary_value = rotary_value;
        Lcd_Chr(1,1, ' ');
        Lcd_Chr(2,1, '>');
      }
      else
      {
        // return to main menu.
        is_in_menu = TRUE;
        rotary_value = menu_pos;
        last_rotary_value = rotary_value + 1;
        Lcd_Chr(2,1, ' ');
        Lcd_Chr(1,1, '>');
      }
    }
    
    // enable global interrupts to capture the rotary encoder updates.
    INTCON |= 0x80;
    portb_state = (PORTB & 0xFC);
    Delay_ms(20);

    // check for menu screen timeout.
    if(timeout_counter > 0)
    {
      if((--timeout_counter) == 0x00)
      {
        timeout_counter = TIMEOUT_LIMIT;
        break;
      }
    }
  }

  // assign volume level to the global counter
  rotary_value = vol_lvl;
  last_rotary_value = rotary_value;
  return;
}

void init_system()
{
  // init CPU and peripheral registers.
  ADCON1 |= 0x0F;
  TRISC = 0x00;
  PORTC = 0x00;
  TRISB = 0x1F;
  INTCON3 = 0x04;
  INTCON2 = 0x40;
  INTCON = 0xD0;
  PORTB = 0xA0;

  // 1602 LCD initialization.
  Delay_ms(100);
  Lcd_Init();
  Lcd_Cmd(_LCD_CLEAR);
  Lcd_Cmd(_LCD_CURSOR_OFF);

  // I2C bus initialization with 100Kbits/sec.
  I2C1_Init(100000);
  Delay_ms(100);
  
  // assign tone and volume values to sound processors.
  if((PORTC & 0x04) == 0x04)
  {
    // call for system memory reset.
    reset_system();
    EEPROM_Write(0x00, vol_lvl);
    EEPROM_Write(0x01, bass_lvl);
    EEPROM_Write(0x02, trib_lvl);
    EEPROM_Write(0x03, loudness);
    EEPROM_Write(0x04, bal_lvl);
    EEPROM_Write(0x05, bass_add);
    EEPROM_Write(0x06, base_gain);
    EEPROM_Write(0x07, base_param);
  }
  else
  {
    // recall settings from EEPROM.
    vol_lvl = EEPROM_Read(0x00);
    bass_lvl = EEPROM_Read(0x01);
    trib_lvl = EEPROM_Read(0x02);
    loudness = EEPROM_Read(0x03);
    bal_lvl = EEPROM_Read(0x04);
    bass_add = EEPROM_Read(0x05);
    base_gain = EEPROM_Read(0x06);
    base_param = EEPROM_Read(0x07);
    recall_system();
  }
  
  return;
}

void reset_system()
{
  // set master volume level to zero.
  update_tone_control(DEFAULT_VOLUME);
  // set speaker mute to off, and activate left and right channels.
  update_tone_control(0x80);
  asm nop;
  update_tone_control(0xA0);
  // enable loudness effect.
  update_tone_control(0x40);
  // set bass and tribble to 0dB level.
  update_tone_control(0x6F);
  asm nop;
  update_tone_control(0x7F);
  // set bass processor to default level.
  Delay_ms(350);
  update_bass_processor(0x00);
  return;
}

void recall_system()
{
  update_tone_control(0x40 | loudness);
  asm nop;
  update_tone_control(0x70 | ((trib_lvl > 7) ? (trib_lvl & 0x08) | ((~trib_lvl) & 0x07) : trib_lvl));
  asm nop;
  update_tone_control(0x60 | ((bass_lvl > 7) ? (bass_lvl & 0x08) | ((~bass_lvl) & 0x07) : bass_lvl));
  asm nop;
  switch(bal_lvl)
  {
    // activate left speaker
    case 0:
      update_tone_control(0x80);
      asm nop;
      update_tone_control(0xBF);
      break;
    // activate right speaker.
    case 2:
      update_tone_control(0xA0);
      asm nop;
      update_tone_control(0x9F);
      break;
    // activate both speakers.
    default:
      update_tone_control(0xA0);
      asm nop;
      update_tone_control(0x80);
  }
  
  Delay_ms(350);
  
  update_bass_processor(base_param | bass_add | (base_gain << 3));
  asm nop;
  update_tone_control(vol_lvl);
  return;
}

void interrupt()
{
  INTCON &= 0x7F;

  // handle rotary encoder movement on phase 1.
  if((INTCON & 0x02) > 0x00)
  {
    if(((PORTB & 0x01) << 1) == (PORTB & 0x02))
      rotary_value++;
    else
      rotary_value--;
  }

  // handle rotary encoder movement on phase 2.
  if((INTCON3 & 0x01) > 0x01)
  {
    if(((PORTB & 0x01) << 1) == (PORTB & 0x02))
      rotary_value--;
    else
      rotary_value++;
  }

  // reset signled registers.
  INTCON &= 0xFD;
  INTCON3 &= 0xFE;
  asm nop;
  INTCON |= 0x80;
  return;
}